#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include "blitmenu.h"
#include "ir.h"

void error_exit(int rc,char *msg) {
	perror(msg);
	exit(rc);
}

struct rawImage *newImage(int width, int height, int fd) {

	struct rawImage	*img;
	char		*data;

	img = (struct rawImage *)calloc(1,sizeof(struct rawImage));
	img->width=width;
	img->height=height;
	data = (char *)malloc(width*height);
	if ( fd == -1 )
		memset(data,0,width*height);
	else {
		if ( read(fd,data,width*height) == -1 ) {
			free(data);
			free(img);
			return NULL;
		}
	}
	img->data=data;
	return img;
}

int freeImage(struct rawImage *img) {
	free(img->data);
	free(img);
}

int freeMeList(struct menuItem *me) {

	while (me->next != NULL) {
		me->prev->next=NULL;
		me=me->next;
		freeImage(me->img);
		free(me->prev);
	}
	freeImage(me->img);
	free(me);
}

int display_on(int fd) {
	ioctl((fd), _IOW('d', 1, int), (1));
}

int display_off(int fd) {
	ioctl((fd), _IOW('d', 1, int), (0));
}

int blitImage(int fd, caddr_t dmap, struct rawImage *img) {
	int i,j=0;
	char cLeft = 0, cRight = 0;
	
	for( i = 0; i < 4096; i+=2, ++j )
	{
		cLeft = *(char *)((img->data)+i) >> 6;
		cRight = *(char *)((img->data)+i+1) >> 6;
		dmap[ j ] = (cRight << 4) | cLeft;
	}
	ioctl((fd), _IO('d', 0));
}

int pasteImage(struct rawImage *dstimg, struct rawImage *srcimg, int x, int y) {

	int dstw, dsth, srcw, srch;
	int i,j;

	dstw=dstimg->width;
	dsth=dstimg->height;
	srcw=srcimg->width;
	srch=srcimg->height;
	for(i=0;i<srcw;i++)
		for(j=0;j<srch;j++) {
			if ( (i+x < dstw) && (j+y < dsth)
				&& (i+x >= 0) && (j+y >= 0) )
				*(char *)(dstimg->data+((j+y)*dstw)+i+x)=
					*(char *)(srcimg->data+(j*srcw)+i);
		}
	
	
	
}

struct rawImage *createTextImage(struct rawImage *asciiImage,char *text) {
	char *t;
	char c[1];
	int i,j=0;
	int s,w,wtot,h=0;
	int sIndex,dIndex=0;
	struct rawImage	*img;

	s=strlen(text);
	wtot=asciiImage->width;
	w=wtot / 95;
	h=asciiImage->height;
	img=newImage( w*s, h, -1);
	
	for( t = text; t < (s + text); t++ ) {
		strncpy(c, t, 1);
		sIndex=((int)*c - 32)*w;
		if ( sIndex < 0 || sIndex > (94*w) )
			sIndex=0;
		for (i=0; i < h; i++) {
			for (j=0; j < w; j++)
				*((char *)
				img->data+dIndex+j+(i*w*s))=
					*((char *)
					asciiImage->data+sIndex+j+(i*wtot));
		}
		dIndex+=w;
	}
	return img;
}

int clearImage(struct rawImage *img) {
	memset(img->data,0,img->width * img->height);
	}


struct rawImage *lowerImage(struct rawImage *src,int drop) {
	
	int i;
	struct rawImage	*dst;

	dst=newImage( src->width, src->height, -1);
	for (i=0; i < (dst->width * dst->height); i++) {
		if ( *((char *)src->data + i) > (drop*64) ) 
			*((char *)dst->data + i) =
				*((char *)src->data + i) - (drop*64);
		else
			*((char *)dst->data + i) = 0;
	}
			
}

int main(int argc, char *argv[])
{
	int irfd, disfd, ascfd, cfgfd;
	FILE *cfgfp;
	ssize_t rts;
	long ircode;
	int leftpressed=0;
	int rightpressed=0;
	int bottompressed=0;
	int updatescreen=1;
	int scroll_left=0;
	int scroll_right=0;
	caddr_t dmap;
	int i = 0, j = 0;
	char s[256],*title,*command;
	char *rawfile;
	int rawwidth,rawheight;
	struct rawImage *asciiImage, *disImage;
	struct rawImage *pprevImg, *prevImg, *selImg, *nextImg, *nnextImg;
	struct menuItem *me, *firstme, *prevme;

	if ( argc < 2 ) {
		fprintf(stderr,
			"usage: %s <configfile>\n",argv[0]);
		exit(1);
	}
	irfd = open("/dev/ir", O_RDONLY|O_NONBLOCK);
	if ( irfd == -1 ) error_exit(1,"Opening IR device.");
	
	disfd = open("/dev/display", O_RDWR);
	if ( disfd == -1 ) error_exit(1,"Opening display");

	cfgfp = fopen(argv[1], "r");
	if ( cfgfp == NULL ) error_exit(1,argv[2]);

	if (fgets(s, sizeof(s),cfgfp) != NULL ) {
		rawfile=(char *)strdup(strtok(s,":"));
		rawwidth=atoi(strtok(NULL,":")) * 95;
		rawheight=atoi(strtok(NULL,"\n"));
		}
	else error_exit(1,"incomplete configfile");
	
	ascfd = open(rawfile, O_RDONLY);
	if ( ascfd == -1 ) error_exit(1,rawfile);
	asciiImage=newImage( rawwidth, rawheight, ascfd);
	close(ascfd);
	free(rawfile);

	me=NULL;firstme=NULL;prevme=NULL;
	while (fgets(s, sizeof(s),cfgfp) != NULL ) {
		title = (char *)strdup(strtok(s,":"));
		command = (char *)strdup(strtok(NULL,"\n"));
		prevme=me;
		me = (struct menuItem *)malloc(sizeof(struct menuItem));
		me->img=createTextImage(asciiImage,title);
		me->title=title;
		me->command=command;
		me->next=NULL;
		me->prev=prevme;
		if ( prevme != NULL ) prevme->next=me;
		if ( firstme == NULL ) firstme=me;
		}
	fclose(cfgfp);

	me->next=firstme;
	firstme->prev=me;
	me=firstme;
	

	dmap = mmap(0, 2048, PROT_READ | PROT_WRITE, MAP_SHARED, disfd, 0);
	if (dmap == (caddr_t) -1)
		return 2;

	disImage=newImage(128,32,-1);
	for (i=0; i<100; i++) 
		rts = read(irfd, (char *) &ircode, 4);
	display_on(disfd);
	i=0;j=0;
	for (;;) {

		rts = read(irfd, (char *) &ircode, 4);

		if (rts == 0 ) {
    			perror("Finished reading from IR device.");
			close(irfd);
			if (munmap(dmap, 2048) == -1 ||
				close(disfd) == -1)
					exit(3);
			exit(0);
		}
		switch (ircode) {

		case IR_TOP_BUTTON_RELEASED:
			break; /* do nothing for releases */
		case IR_TOP_BUTTON_PRESSED: {
			display_off(disfd);
			exit(0);
			}
		case IR_BOTTOM_BUTTON_PRESSED: {
			bottompressed=1;
			break;
			}
		case IR_BOTTOM_BUTTON_RELEASED:
			if ( bottompressed ) {
				close(irfd);
				if (munmap(dmap, 2048) == -1 ||
					close(disfd) == -1)
						exit(3);
				fprintf(stdout,"%s",me->command);
				exit(0);
				}
		case IR_LEFT_BUTTON_PRESSED: {
			leftpressed=1;
			break;
			}
		case IR_LEFT_BUTTON_RELEASED: 
			if ( leftpressed ) {
				me=me->prev;
				leftpressed=0;
				updatescreen=1;
				scroll_left=1;
				break;
			}
		case IR_RIGHT_BUTTON_PRESSED: {
			rightpressed=1;
			break;
			}
		case IR_RIGHT_BUTTON_RELEASED:
			if ( rightpressed ) {
				me=me->next;
				rightpressed=0;
				updatescreen=1;
				scroll_right=1;
				break;
			}
		}

		if (updatescreen) {
			prevImg=lowerImage(me->prev->img,2);
			nextImg=lowerImage(me->next->img,2);
		if(scroll_left) {
			selImg=lowerImage(me->img,2);
			nnextImg=lowerImage(me->next->next->img,2);
		
			for (i=0; i<=(me->img->height+1); i+=2) {
				clearImage(disImage);
				pasteImage(disImage,prevImg,-32,i-11);
				pasteImage(disImage,selImg,0,i);
				pasteImage(disImage,nextImg,32,i+11);
				pasteImage(disImage,nnextImg,64,i+22);
				blitImage(disfd,dmap,disImage);
				usleep(10000);
			}
			for (i=0; i<=32; i+=4) {
				clearImage(disImage);
				pasteImage(disImage,prevImg,i-32,0);
				pasteImage(disImage,selImg,i,11);
				pasteImage(disImage,nextImg,i+32,22);
				blitImage(disfd,dmap,disImage);
				usleep(10000);
			}
			scroll_left=0;
			freeImage(selImg);
			freeImage(nnextImg);
		}
		if(scroll_right) {
			pprevImg=lowerImage(me->prev->prev->img,1);
			selImg=lowerImage(me->img,1);
		
			for (i=0; i<=(me->img->height+1); i+=2) {
				clearImage(disImage);
				pasteImage(disImage,pprevImg,0,-i);
				pasteImage(disImage,prevImg,32,11-i);
				pasteImage(disImage,selImg,64,22-i);
				pasteImage(disImage,nextImg,96,33-i);
				blitImage(disfd,dmap,disImage);
				usleep(10000);
			}
			for (i=0; i<=32; i+=4) {
				clearImage(disImage);
				pasteImage(disImage,prevImg,32-i,0);
				pasteImage(disImage,selImg,64-i,11);
				pasteImage(disImage,nextImg,96-i,22);
				blitImage(disfd,dmap,disImage);
				usleep(10000);
			}
			scroll_right=0;
			freeImage(pprevImg);
			freeImage(selImg);
		}
			pasteImage(disImage,prevImg,0,0);
			pasteImage(disImage,me->img,32,11);
			pasteImage(disImage,nextImg,64,22);
			blitImage(disfd,dmap,disImage);
			usleep(10000);
			updatescreen=0;
			freeImage(prevImg);
			freeImage(nextImg);
		}
		usleep(10000);
	}

	freeImage(asciiImage);
	freeImage(disImage);
	freeMeList(me);
	close(irfd);
	if (munmap(dmap, 2048) == -1 || close(disfd) == -1)
		return 3;

	return 0;
}
